/* qtalk.c */

/* Change History ------------------------------------------------------------
*
*  version  date     by  description
*
*  04.01.01 01/14/93 jpc added -h option to set tcdropline duration
*
* End Change History --------------------------------------------------------
*/


#define STEAL_ALLOWED "S"

#ifdef __USAGE
%C	[options]*
Options:
 -c <cmd>        Command char.
 -d <del>        Delete char.
 -e              Local echo on.
 -E              Suppress FF codes (qnxs).
 -i              Input flow control on.
 -k <kick>       Kick char.
 -l <logfile>    Path to log file.
 -m <port>       Device to use.
 -o <options>    Transfer protocol options.
 -p <pause>      Pause char.
 -P              Strip parity.
 -s <system>     System to call.
 -h <millisec>   Disconnect duration.
Note:
 For HELP, press Ctrl-A followed by ? within Qtalk.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dev.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/wait.h>
#include <termios.h>
#include <malloc.h>
#include <signal.h>
#include <string.h>
#include <setjmp.h>
#include <fcntl.h>
#include <process.h>
#include <errno.h>

unsigned a_dev_mode(int fd, unsigned mode, unsigned mask);

#define old_mode()	{a_dev_mode(kbd, okbd_mode, _DEV_MODES);\
					 a_dev_mode(mdm, omdm_mode, _DEV_MODES);}

#define new_mode()	{a_dev_mode(kbd,  kbd_mode, _DEV_MODES);\
					 a_dev_mode(mdm,  mdm_mode, _DEV_MODES);}

#define VERSION				"[4.1X]"
#define MDM_VERSION			"q10"
#define BUFFER_SIZE			4000
#define DEFAULT_CMD_CHAR	0x01
#define ENQ					0x05
#define SYN					0x16
#define ESC					0x1B
#define RUB					0x7F
#define DIAL_DIR			"/etc/config/dial.dir"
#define LINE_LENGTH			256
#define OPT_LENGTH			LINE_LENGTH

int						mdm, scr, kbd, log_file = -1;
int						disconnect = 500;
int						mdmflag, kbdflag, qnxs;
pid_t					pid;
unsigned				mproxy = 0, kproxy = 1, vmproxy = 0, vkproxy = 1;

sigjmp_buf				jmpbuf;
unsigned				kbd_mode, mdm_mode, okbd_mode, omdm_mode;
struct	_dev_info_entry	kbd_entry, mdm_entry;

char		*mname,
			mtbuf[32],
			buf[ BUFFER_SIZE ],
			buf2[ BUFFER_SIZE ],
			*sys,
			cmd_char,
			del_char,
			pause_char,
			kick_char,
			protocol[ LINE_LENGTH + 1 ],
			transfer_options[ OPT_LENGTH + 1 ],
			line[ LINE_LENGTH + 1 ];
int			num_packets;
long		num_chars;
int			show_stats,
			has_timeouts,
			local_echo,
			strip_parity,
			numbrks;


/*-------------------------------------------------- IO COVER FNS
        They abort when they get 'significant' errors.
*/

void a_abrt(int err, char *fn, int fd)
{
	fprintf(stderr,"\r\nqtalk: (%s) %s\r\n",fn,strerror(err));
	if (mdm!=-1 && fd!=mdm) tcdropline(mdm,500);
    if (kbd!=-1 && fd!=kbd) dev_mode(kbd, okbd_mode, _DEV_MODES);
	if (mdm!=-1 && fd!=mdm) dev_mode(mdm, omdm_mode, _DEV_MODES);

	if (err==ESRCH || err==EBADF) {
		if (fd==mdm) fprintf(stderr,"qtalk: May have lost network VC to modem device.\r\n");
	}

	exit (EXIT_FAILURE);
}

ssize_t a_read(int fildes, void *buf, size_t nbyte)
{
	ssize_t rc;	

	if ((rc = read(fildes,buf,nbyte))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:		/* Old MacDonald had a farm, EI EIO... */
				break;
			default:		/* Old MacDonald bought the farm... */
				a_abrt(errno,"read",fildes);
		}
	}	
	return rc;
}

ssize_t a_write(int fildes, const void *buf, size_t nbyte)
{
	ssize_t rc;

	if ((rc = write(fildes,buf,nbyte))==-1) {
		switch (errno) {
			case EINTR:
			case EIO:
				rc=-1;
				break;
			default:
				a_abrt(errno,"write",fildes);
		}
	}

	return rc;
}

int a_dev_ischars(int fd)
{
	int rc;

	if ((rc = dev_ischars(fd))==-1)
		a_abrt(errno,"dev_ischars",fd);
	return rc;                      
}


int a_dev_arm(int fd, pid_t proxy, unsigned events)
{
	int rc;

	if ((rc = dev_arm(fd,proxy,events))==-1) 
		a_abrt(errno,"dev_arm",fd);

	return rc;
}


unsigned a_dev_mode(int fd, unsigned mode, unsigned mask)
{
	unsigned rc;

	if ((rc = dev_mode(fd,mode,mask))==-1) 
		a_abrt(errno,"dev_mode",fd);
	return rc;
}

int a_dev_read(int fd, void *buf, unsigned n, unsigned min, unsigned time,
               unsigned timeout, pid_t proxy, int *armed)
{
	int rc;

	if ((rc = dev_read(fd,buf,n,min,time,timeout,proxy,armed))==-1) {
		switch (rc) {
			case EINTR:
			case EIO:
				break;
			default:
				a_abrt(errno,"dev_read",fd);
		}
	}
	return rc;
}

int a_tcdrain(int fildes)				
{
	int rc;

	if ((rc = tcdrain(fildes))==-1) {
		switch(rc) {
			case EINTR:
				break;
			default:
				a_abrt(errno,"tcdrain",fildes);
		}
	}
	return rc;
}

/* end of IO cover fns ------------------------------------------------ */				


char *get_name( char *prompt )
{
	static char	name[ LINE_LENGTH + 1 ];
	int			i;

	old_mode();
	a_write( kbd, prompt, strlen( prompt ) );
	fgets( &name, LINE_LENGTH, stdin );
	new_mode();

	i = strlen(name)-1;
	if ( name[i] == '\n' )
		name[i] = '\0';

	return( &name );
}

void gotsig( int sig )
{
	if ( sig == SIGINT ) {
		tcsendbreak( mdm, 300 );
		siglongjmp( &jmpbuf, 0 );
	}
	old_mode();
	exit( sig );
}

void mdm_handler( char *bp, int size, int *armed )
{
	static int			 numsyns,
						 esc_seen;
	char				*p;
	int					 n;


	put_console( bp, size);
	/*
	 * post-process the line
	 */
	p = bp;
	for(n = 0; n < size; ++n) {
		/*
		 * Respond to 2 consecutively received SYN characters
		 */
		if(*p == SYN) {
			if(++numsyns == 2 ) {
				/*
				 *	After 2 in a row, invoke QCP.
				 */

				/* turn off modem proxies */
				a_dev_arm(mdm,-1,_DEV_EVENT_INPUT);

				/* flush any pending modem proxies */
				while (Creceive(mproxy,bp,0)!=-1);

				/* read and discard any pending modem data */
				while (n=a_dev_ischars(mdm)) {
					if (n==-1) break;
					/* have BUFFER_SIZE to play with */
					a_read(mdm,bp,n>BUFFER_SIZE?BUFFER_SIZE:n);
				}
				
				sprintf( line, "%s %s re %s", protocol, mname, transfer_options );
				old_mode();
				system( line );
				new_mode();
				numsyns = 0;

				/* turn modem proxies back on */
				a_dev_arm(mdm,vmproxy,_DEV_EVENT_INPUT);

				*armed = 0;	/* ready to go back to receive */

				errno=0;
				break;	/* ignore the rest of the line */
			}
		} else numsyns = 0;
		/*
		 * Write back ID after <Esc><Enq>
		 */
		if(esc_seen && *p == ENQ) {
			a_write( mdm, MDM_VERSION, 3 );
		}
		if(*p == ESC) esc_seen = 1;
		else esc_seen = 0;
		++p;
	}
}

int put_console( char *buf, int n )
{
	int	 i;
	char *p = buf;

	if (strip_parity)	for(i=0; i<n; ++p,++i) *p &= 0x7F;

	if (qnxs) {
		char *buf2=buf;	/* current print position */
		char *replace_with;
		int  numtoprint=0;

		for (i=0;i<n;++i,++numtoprint) {
			switch(*p) {
				case 0xa6:	replace_with = "c"; break; /* right arrow */
				case 0xa4:	replace_with = "d"; break; /* left arrow */
				case 0xa1:	replace_with = "a"; break; /* up arrow */
				case 0xa9:	replace_with = "b"; break; /* down arrow */
				case 0xa0:	replace_with = "H"; break; /* home key */
				default: replace_with = NULL; break;
			}

			if (replace_with!=NULL) {
				a_write(kbd,buf2,numtoprint);	/* write line to this point */
				if (log_file != -1) put_logfile( buf2, numtoprint );
				buf2+=numtoprint+1;			/* increment 'write from' position */
				a_write(kbd,replace_with,2);	/* write replacement for ctrl sequence */
				if (log_file != -1) put_logfile( replace_with, 2 );
				numtoprint=-1;				/* will be inc'ed to 0 as we loop */
			}
		}
		if (numtoprint) a_write(kbd,buf2,numtoprint);
		if (log_file != -1) put_logfile( buf2, numtoprint );
	} else {
		a_write( kbd, buf, n );
		if (log_file != -1) put_logfile( buf, n );
	}

	return( n );
}

int put_logfile( char *buf, int n )
{
	register char	*p1,
					*p2,
					*endp;
	register int	 count = 0;

	for( p1 = p2 = buf, endp = p1 + n; p1 < endp; ++p1 ) {
		if(*p1 != '\r') {							/* Strip CRs */
			*p2++ = *p1;
			++count;
		}
	}

	write( log_file, buf, count );

	return  n;
}

int dial( char *system )
{
	FILE	*fp;
	char	*p;
	int		n;

	if (!(fp=fopen(DIAL_DIR,"r")))	return 0;

	if ( strcmp( system, "?" ) == 0 ) {
		while( 1 ) {
			p = fgets( line, LINE_LENGTH, fp );

			if(p == NULL) break;
			n = strlen(p);

			if(n<1)  break;
			line[n-1] = '\0';
			fprintf( stderr, "%s\r\n", line );
		}
		fclose( fp );
		return 2;
	} else while( 1 ) {
		p = fgets( line, LINE_LENGTH, fp );
		if(p == NULL) break;
		n = strlen( p );
		if( n < 1) continue;
		line[n - 1] = '\0';
		for(p=line; *p; ++p) if (*p==' '||*p=='\t') break;

		if (*p == '\0') continue;		/*	Check for a dialing string	*/

		*p++ = '\0';
		if ( strcmp( line, system ) == 0 ) {
			/*
			 *	Skip leading whitespace before the dialing string
			 */
			while(*p && (*p==' ' || *p=='\t')) ++p;

			if (*p == '\0') break;	/*	Check for null dialing string	*/

			a_write( mdm, p, strlen( p ) );
			a_write( mdm, "\r", 1 );
			fclose( fp );
			return 1;
		}
	} /* else while (1) */

	fprintf( stderr, "Unable to dial '%s': no dialing entry\r\n", system );
	fclose( fp );
	return 0;
}

void kbd_handler( char *bp, int size )
{
	char	*p, *p1;
	unsigned char c;
	int		n=0;
	int command_flag = 0;
	int i;

	p = p1 = bp;

	for(i = 0; i < size; ++i) {
		c = *p++;
		if(c == cmd_char) {
			++command_flag;
			break; /* Ignore the rest of the data */
		}
		if (qnxs && c==0xFF) continue;
		if (del_char && c==RUB)	c=del_char;
		*p1++ = c;
		++n;
	}

	if(n > 0) {
		a_write( mdm, bp, n );
		if (local_echo) put_console( bp, n );
	}

	if(command_flag) command();

}


void command(void)
{
	register char	*p;
	char			 c;

	printf( "\r\nQtalk %s\r\n",VERSION);
	printf( "Local echo %s\r\n",local_echo?"enabled":"disabled");
	printf( "Logging    %s\r\n",log_file!=-1?"enabled":"disabled");
	printf( "Parity bit %s\r\n",strip_parity?"stripped":"displayed");

again:
	printf( "\r\nCommand(?): ");
	fflush( stdout );
	read( kbd, &c, 1 );
	if (isalnum(c) || c=='?') putc( c, stdout );

	printf( "\r\n" );
	switch(tolower(c)) {
		case 'b':	tcsendbreak( mdm, 300 );	break;

		case 'c':	getcwd( &line, LINE_LENGTH );
					printf( "Current Directory is: %s\r\n", &line);
					p = get_name( "Directory? " );
					if (strlen(p) > 0) {
						if ( chdir( p ) != 0 )
							printf( "Unable to change to directory %s\r\n", p );
					}
					break;

		case 'd':	do {
						p = get_name("System? (? for list) ");
						if (*p=='\0') break;
					} while( dial( p ) == 2 );
					break;

		case 'e':	local_echo=!local_echo;
					printf("Local echo %s\r\n",local_echo?"enabled":"disabled");
					break;

		case 'l':	if ( log_file != -1 ) {	/*	toggle log status	*/
						close( log_file );
						log_file = -1;
						printf( "Log file closed\r\n");
					} else log_file=open(get_name("Log file?"),O_WRONLY|O_APPEND|O_CREAT,0666);
					break;

		case 'o':	printf("Current transfer options are \"%s\"\r\n",transfer_options );
					p = get_name( "Transfer options: " );
					if(strlen(p)>0) strcpy( transfer_options, p );
					break;

		case 't':	printf("Current protocol is \"%s\"\r\n",protocol);
					p = get_name( "Protocol: " );
					if (strlen(p)>0) strcpy(&protocol[0],p);
					break;

		case 'p':	strip_parity=!strip_parity;
					printf("Parity bit %s\r\n",strip_parity?"stripped":"displayed");
					break;

		case 'r':	sprintf(line,"%s %s re %s",protocol,mname,transfer_options);
					old_mode();
					system( line );
					new_mode();
					break;

		case 's':	sprintf(line,"%s %s se %s %s",protocol,mname,transfer_options,
							 get_name( "Send file(s)? " ) );
					old_mode();
					system( line );
					new_mode();
					break;

		case 'u':	mdm_write( 1 );	break;

		case 'w':	mdm_write( 0 );	break;

		case 'h':	tcdropline(mdm, disconnect);break;

		case 'q':	tcdropline( mdm, disconnect );	/* fall thru */
		case 'x':	old_mode();
					exit(EXIT_SUCCESS);

		case '2':	if (qnxs) {
						printf("qnxs mode disabled. (Leading FFs will be xmitted)\r\n");
						qnxs=0;
					} else {
						printf("qnxs mode enabled. (Leading FFs will be stripped)\r\n");
						qnxs++;
					}
					break;

		case '!':	p = get_name("Enter command: ");
					if(strlen(p) > 0) {
						old_mode();
						system( p );
						new_mode();
					}
					break;
		default:	if( c == cmd_char ) {
						a_write( mdm, &c, 1 );
						break;
					}
					if (isalnum(c)||c=='?') {
						printf( "%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r%s\n",
								"b - Send Break             q - Quit & Hangup",
								"c - Change directory       r - Receive a file",
								"d - Dial System            s - Send File",
								"e - Toggle Local Echo      t - Specify transfer protocol",
								"h - Hangup                 u - Upload Text",
								"l - Log File On/Off        w - Write File to Modem",
								"o - Set transfer options   x - Exit w/o Hangup",
								"p - Toggle Parity Filter   ! - Escape to shell",
								"2 - Toggle qnxs mode (To talk to QNX2.x tset qnxs.)"
						);
						goto again;
					}
	}
}

/* this is used only by -u and -w options. What the heck is this for? */

void mdm_write( int upload_flag )
{
	int		fd, n, i;
	char	c, kick_seen;

	if ( ( fd = open( get_name( "Upload file? " ), O_RDONLY ) ) == -1 )
		return;

	while( a_read( fd, &c, 1 ) == 1 ) {
		if ( local_echo ) {
			if (c=='\n') put_console("\r\n", 2);
			else put_console( &c, 1 );
		}

		if (c=='\n') c='\r';

		a_write( mdm, &c, 1 );

		kick_seen = (upload_flag && !local_echo	|| pause_char && kick_char
					 && c==pause_char)? 0 : 1 ;
		/*
		 *	Only flush if you are about to wait for echo or kick.
		 *  99% of the time a <W>rite won't go through this code.
		 */
		if (!kick_seen) a_tcdrain(mdm);

		while(!kick_seen) {
#ifdef DEVARM
			if ( !mdmflag ) {
				a_dev_arm( mdm, mproxy, _DEV_EVENT_INPUT );
				mdmflag = 1;
			}
			if ( !kbdflag ) {
				a_dev_arm( kbd, kproxy, _DEV_EVENT_INPUT );
				kbdflag = 1;
			}
			pid = Receive( 0, 0, 0 );
#else
			for( ;; ) {
				if ( a_dev_ischars( mdm ) ) {
					pid = mproxy;
					break;
				}
				else if ( a_dev_ischars( kbd ) ) {
					pid = kproxy;
					break;
				}
			}
#endif

			if ( pid == kproxy ) {
				printf( "Upload aborted.\r\n" );
				goto end;
			} else if ( pid == mproxy ) {
				n = a_read( mdm, &buf, a_dev_ischars( mdm ) );
				put_console( &buf, n );
				if ( pause_char  &&  kick_char  &&  c == pause_char ) {
					for( kick_seen = i = 0; !kick_seen  &&  i < n; ++i ) {
						if (buf[i]==kick_char)	kick_seen = 1;
					}
				} else kick_seen = 1;
			} else {
				fprintf(stderr, "qtalk: Received an unexpected message from pid %d\r\n", pid );
			}
		}
	}

	a_tcdrain( mdm );
	printf( "Upload completed\r\l");
	fflush( stdout );
end:
	close( fd );
}


main( int argc, char *argv[] )
{
	unsigned			 iflow = 0,
#ifdef STEAL_ALLOWED
						 steal = 0,
#endif
						 opt,
						 n;
	int					 armed;
	char				*cp;
	extern void			 gotsig( );

	strcpy( protocol, "qcp" );
	mname = "/dev/ser1";
	cmd_char = DEFAULT_CMD_CHAR;

	while((opt=getopt(argc,argv,"h:m:l:iePEo:c:d:k:p:s:" STEAL_ALLOWED ))!=-1) {
		switch(opt) {
 			case 'm':	mname = optarg;						break;
 			case 'l':	log_file=open(optarg,O_WRONLY|O_APPEND|O_CREAT,0666);
						break;
 			case 'i':	iflow = 1;						 	break;
 			case 'e':	++local_echo;					 	break;
 			case 'P':	++strip_parity;					 	break;
 			case 'E':	++qnxs;								break;
 			case 'o':	strcpy( transfer_options, optarg ); break;
 			case 'c':	cmd_char = (char)atoh( optarg ); 	break;
 			case 'd':	del_char = (char)atoh( optarg ); 	break;
 			case 'k':	kick_char = (char)atoh( optarg );	break;
 			case 'p':	pause_char = (char)atoh( optarg ); 	break;
 			case 's':	sys = optarg;					 	break;
#ifdef STEAL_ALLOWED
 			case 'S':	steal = 1;						 	break;
#endif
			case 'h':	disconnect = atoi( optarg );		break;
			default:	exit(EXIT_FAILURE);					break;
		}
	}

	if (optind<argc) sys=argv[optind];

	kbd = 0;
	scr = 1;
	mdm = open(mname,O_RDWR);
	if (mdm==-1) {
		fprintf(stderr, "qtalk: Unable to open device %s\n", mname );
		exit(EXIT_FAILURE);
	}

	mname = strcpy(&mtbuf[0], ttyname(mdm));
	printf( "Using modem %s\n", mname );

	dev_info( kbd, &kbd_entry );
	dev_info( mdm, &mdm_entry );

	if(mdm_entry.open_count > 1) {
		char nodebuf[10];
#ifndef STEAL_ALLOWED
		fprintf(stderr, "qtalk: Device %s in use.\n",mname);
		exit(EXIT_FAILURE);
#else
		if(!steal) {
			fprintf(stderr, "qtalk: Device %s in use. Use -S to try and steal.\n", mname );
			exit(EXIT_FAILURE);
		}

		for(cp=mdm_entry.tty_name+2,n=0;*cp>='0'&&*cp<='9';++cp,++n)
			nodebuf[n] = *cp;

		nodebuf[n] = '\0';

		n = spawnl(P_WAIT,	"/bin/slay", "slay",
							"-S",
							"-Q",
							"-s",
							"quit",
							"-n", nodebuf,
							"-t", cp,
							"modem",
							NULL);

		if(n < 1 || WEXITSTATUS(n) != 1) {
			fprintf(stderr, "qtalk: Unable to slay modem to steal device %s.\n", mname );
			exit(EXIT_FAILURE);
		}
#endif
	}

	if ( ( mproxy = qnx_proxy_attach( 0, 0, 0, -1 ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create modem proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( vmproxy = qnx_proxy_rem_attach( mdm_entry.nid, mproxy ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create remote modem proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( kproxy = qnx_proxy_attach( 0, 0, 0, -1 ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create keyboard proxy\n" );
		exit(EXIT_FAILURE);
	}

	if ( ( vkproxy = qnx_proxy_rem_attach( kbd_entry.nid, kproxy ) ) == -1 ) {
		fprintf(stderr, "qtalk: Unable to create remote keyboard proxy\n" );
	exit(EXIT_FAILURE);
	}

	signal( SIGINT, &gotsig );
	signal( SIGQUIT, &gotsig );
	signal( SIGTERM, &gotsig );

	/* Set the keyboard and modem completely RAW */
	kbd_mode = 0;
	okbd_mode = a_dev_mode( kbd, kbd_mode, _DEV_MODES);

	mdm_mode = 0;
	omdm_mode = a_dev_mode( mdm, mdm_mode, _DEV_MODES);

	/* flush all the crud there to start with */
	tcflush( mdm, TCIOFLUSH );
	tcflush( kbd, TCIOFLUSH );

	printf( "Qtalk %s\r\n", VERSION );

	if (sys) if (dial(sys)==0) exit(EXIT_FAILURE);

	sigsetjmp( jmpbuf, 1 );	/* If we get a BREAK, send one and return here */

	/*
	 * Arm the proxies by reading zero bytes
	 */
	a_dev_read( kbd, &buf, 0, 1, 0, 0, vkproxy, 0);
	a_dev_read( mdm, &buf, 0, 1, 0, 0, vmproxy, 0);

	for( ;; ) {
		pid = Receive( 0, &buf, 0 );
		if ( pid == mproxy ) {
			armed = 0;
			while(!armed) {
				n = a_dev_read( mdm, &buf[0], BUFFER_SIZE, 1, 0, 0, vmproxy, &armed);
				strncpy(buf2,buf,n+1);
				buf2[n+1]=0;	
				if (n==-1) perror("dev_read(mdm...)");
				mdm_handler( &buf[0], n, &armed );
			}
		} else if ( pid == kproxy ) {
			armed = 0;
			while(!armed) {
				n = a_dev_read( kbd, &buf[0], BUFFER_SIZE, 1, 0, 0, vkproxy, &armed);
				kbd_handler( &buf[0], n );
			}
		} else Reply(pid,"\131",1);	/* reply ENOSYS */
	}
	exit (EXIT_SUCCESS);
}

